----------------------------------------------------------------------------------
-- POLYTECHNIQUE MONTREAL
-- ELE3311 - Systemes logiques programmables 
-- 
-- Module Name:    ctrl_temp_del 
-- Description:    Niveau hiérarchique supérieur pour
--                 le contrôleur de température pour ROADM
--                 par Pulse Width Modulation (PWM) et affichage
--                 sur DEL
--
-- Additional Comments: 
--
----------------------------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;

-- Uncomment the following library declaration if using
-- arithmetic functions with signed or unsigned values
use IEEE.numeric_std.all;

-- Uncomment the following library declaration if instantiating
-- any Xilinx primitives in this code.
library UNISIM;
use UNISIM.VComponents.all;

entity ctrl_temp_del is
generic (
  SHORT_SIM             : boolean := false;
  HEATERS               : integer := 4;
  PRECISION             : integer := 8
);
port (
  --RST & CLK
  rst_n_i               : in    std_logic;
  clk_i                 : in    std_logic;
      
  prog_i                : in    std_logic;
  next_i                : in    std_logic;
  btn_i                 : in    std_logic_vector(2 downto 0);
  ratio_i               : in    std_logic_vector(7 downto 0);

  del_o                 : out   std_logic_vector(7 downto 0)
);
end ctrl_temp_del;


architecture behavioral of ctrl_temp_del is
-- ###########################################################################
-- Entete de l'architecture
-- ###########################################################################

  signal rst                    : std_logic;
  signal clk                    : std_logic;
  signal sample_8khz            : std_logic;
  signal hb_heartbeat           : std_logic;
  signal next_clk               : std_logic;
  signal prog_clk               : std_logic;
  signal btn_clk                : std_logic_vector(2 downto 0);
  signal dbnc_next              : std_logic;
  signal dbnc_prog              : std_logic;
  signal dbnc_btn               : std_logic_vector(2 downto 0);
  signal addr_cnt_rst           : std_logic;
  signal addr_cnt_inc           : std_logic;
  signal addr_cnt               : std_logic_vector(1 downto 0);
  signal loop_cnt               : std_logic_vector(7 downto 0);
  signal decod_wr_en            : std_logic_vector(HEATERS-1 downto 0);
  signal pwm_heat               : std_logic_vector(HEATERS-1 downto 0);


  component meta_harden	
  generic (
    GENERIC_IO_LOGIC : std_logic := '1'; -- 1=POSITIVE 0=NEGATIVE
    WIDTH            : integer   := 1
  );
  port ( 
    sig_src_i  : in  std_logic_vector(WIDTH-1 downto 0);
    rst_dst_i  : in  std_logic;
    clk_dst_i  : in  std_logic;
    sig_dst_o  : out std_logic_vector(WIDTH-1 downto 0)
  );
  end component;

  component dbnc
  port (
    rst_i                 : in    std_logic;
    clk_i                 : in    std_logic;
    sample_i              : in    std_logic;  -- sample signal should be around 8 kHz
    sig_i                 : in    std_logic;
    dbnc_sig_o            : out   std_logic
  );
  end component;

  component msa_reg
  port (
    rst_i                 : in    std_logic;
    clk_i                 : in    std_logic;
    next_i                : in    std_logic;
    addr_cnt_rst_o        : out   std_logic;
    addr_cnt_inc_o        : out   std_logic
  );
  end component;

  component cnt
  generic (
    WIDTH                 : integer := 8
  );
  port (
    rst_i                 : in    std_logic;
    clk_i                 : in    std_logic;
    dec_i                 : in    std_logic;
    inc_i                 : in    std_logic;
    cnt_o                 : out   std_logic_vector(WIDTH-1 downto 0)
  );
  end component;

  component decod_1p4
  port (
    rst_i                 : in    std_logic;
    clk_i                 : in    std_logic;
    en_i                  : in    std_logic;
    addr_i                : in    std_logic_vector(1 downto 0);
    decod_o               : out   std_logic_vector(3 downto 0)
  );
  end component;

  component pwm
  generic (
    PRECISION            : integer              := 8
  );
  port (
    rst_i                 : in    std_logic;
    clk_i                 : in    std_logic;
    wr_en_i               : in    std_logic;
    ratio_i               : in    std_logic_vector(PRECISION-1 downto 0);
    loop_cnt_i            : in    std_logic_vector(PRECISION-1 downto 0);
    pwm_heat_o            : out   std_logic
  );
  end component;

  component delay_cnt
  generic (
    -- Default is 30M cycles, with a 100MHz clock, which is 300 ms
    -- This is a VHDL 2008 construct, modify FILE_TYPE property in Vivado
    SHORT_SIM             : boolean               := false;
    COUNT_VAL             : unsigned(24 downto 0) := to_unsigned(30000000, 25);
    COUNT_VAL_SHORT       : unsigned(24 downto 0) := to_unsigned(2, 25)
  );
  port (
    rst_i                 : in    std_logic;
    clk_i                 : in    std_logic;
    start_delay_i         : in    std_logic;
    end_delay_o           : out   std_logic
  );
  end component;

  component heartbeat
  generic (
    SHORT_SIM             : boolean := false;
    COUNT_VAL_SHORT       : unsigned(26 downto 0) := resize(d"250", 27);
    COUNT_VAL             : unsigned(26 downto 0) := resize(d"25000000", 27)
    );
  port (
    clk_i                 : in    std_logic;
    heartbeat_o           : out   std_logic
  );
  end component;

-- ###########################################################################
-- Debut de l'architecture
-- ###########################################################################
begin

  ----------------------------------------------------------------------------
  -- Horloge et remise a zero
  -- Instantiation d'un BUFG pour la distribution de l'horloge
  -- Generation du signal d'echantillonnage pour le dbnc
  -- Ajustement de la polarite du reset
  ----------------------------------------------------------------------------
  -- *** A completer ***
  BUFG_inst : BUFG
  port map (
      O => clk,
      I => clk_i
  );

  

  rst      <= not rst_n_i;

  -- On utilise le compteur de delai
  -- Le signal de sortie amorce un nouveau compte
  sample_8khz_inst : delay_cnt
  generic map(
    SHORT_SIM               => SHORT_SIM,
    COUNT_VAL               => to_unsigned(12500, 25),
    COUNT_VAL_SHORT         => to_unsigned(2, 25)
  )
  port map(
    rst_i                   => rst,
    clk_i                   => clk,
    start_delay_i           => sample_8khz,
    end_delay_o             => sample_8khz
  );


  ----------------------------------------------------------------------------
  resynchronisation_inst : meta_harden
  generic map(
    WIDTH                   => 5 
  )
  port map(
    sig_src_i(4)            => next_i,
    sig_src_i(3)            => prog_i,
    sig_src_i(2 downto 0)   => btn_i,
    rst_dst_i               => rst,
    clk_dst_i               => clk,
    sig_dst_o(4)            => next_clk,
    sig_dst_o(3)            => prog_clk,
    sig_dst_o(2 downto 0)   => btn_clk
  );


  ----------------------------------------------------------------------------
  dbnc_next_inst : dbnc
  port map(
    rst_i                   => rst,
    clk_i                   => clk,
    sample_i                => sample_8khz,
    sig_i                   => next_clk,
    dbnc_sig_o              => dbnc_next
  );
  
  dbnc_prog_inst : dbnc
  port map(
    rst_i                   => rst,
    clk_i                   => clk,
    sample_i                => sample_8khz,
    sig_i                   => prog_clk,
    dbnc_sig_o              => dbnc_prog
  );
  
  DBNC_GEN: 
  for W in 0 to 2 generate
    dbnc_btn_inst : dbnc
    port map(
      rst_i                   => rst,
      clk_i                   => clk,
      sample_i                => sample_8khz,
      sig_i                   => btn_clk(W),
      dbnc_sig_o              => dbnc_btn(W)
    );
  end generate DBNC_GEN;


  ----------------------------------------------------------------------------
  msa_reg_inst : msa_reg
  port map(
    rst_i                   => rst,
    clk_i                   => clk,
    next_i                  => dbnc_next,
    addr_cnt_rst_o          => addr_cnt_rst,
    addr_cnt_inc_o          => addr_cnt_inc
  );

  addr_cnt_inst : cnt
  generic map(
    WIDTH                   => 2
  )
  port map(
    clk_i                   => clk,
    rst_i                   => addr_cnt_rst,
    dec_i                   => '0',
    inc_i                   => addr_cnt_inc,
    cnt_o                   => addr_cnt
  );

  ----------------------------------------------------------------------------
  decod_1p4_inst : decod_1p4
  port map(
    clk_i                   => clk,
    rst_i                   => rst,
    en_i                    => dbnc_prog,
    addr_i                  => addr_cnt,
    decod_o                 => decod_wr_en
  );


  ----------------------------------------------------------------------------
  loop_cnt_inst : cnt
  generic map(
    WIDTH                   => PRECISION
  )
  port map(
    clk_i                   => clk,
    rst_i                   => rst,
    dec_i                   => '0',
    inc_i                   => '1',
    cnt_o                   => loop_cnt
  );

  ----------------------------------------------------------------------------
  PWM_GEN: 
  for h in 0 to HEATERS-1 generate
    pwm_inst : pwm
    generic map(
      PRECISION              => PRECISION
    )
    port map(
      rst_i                   => rst,
      clk_i                   => clk,
      wr_en_i                 => decod_wr_en(h),
      ratio_i                 => ratio_i,
      loop_cnt_i              => loop_cnt,
      pwm_heat_o              => pwm_heat(h)
    );
  end generate PWM_GEN;


  ----------------------------------------------------------------------------
  heartbeat_inst : heartbeat
  generic map(
    SHORT_SIM               => SHORT_SIM
  )
  port map(
    clk_i                   => clk,
    heartbeat_o             => hb_heartbeat
  );
  
  ----------------------------------------------------------------------------
  -- Glue logic
  ----------------------------------------------------------------------------

  ----------------------------------------------------------------------------
  -- Assign outputs
  ----------------------------------------------------------------------------
  del_o(7)            <= hb_heartbeat;
  del_o(6 downto 5)   <= addr_cnt;
  del_o(4)            <= '0';             -- Tie low, inactive
  del_o(3 downto 0)   <= pwm_heat;

end behavioral;

